/*
 * Copyright (C) 1998
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <stdlib.h>	/* for getenv(3) */

#include "monitor.h"	/* for macro BINCLASS_{,UN}LOCK() */
#include "compiler.h"


/*
 * Global Variables
 */
bool_t compiler_enabled = TRUE;

int opt_quiet = 0;
int opt_outcode = 0;
int opt_codesize = 0;
int opt_dontcmplvmcls = 0;


/*
 * Local Functions
 */
static void initializeClass(ClassClass *);
static void signalHandler(int sig, void *info, void *uc);
static void freeClass(ClassClass *);
unsigned char *compiledCodePC(JavaFrame *, struct methodblock *);
bool_t pcInCompiledCode(unsigned char *pc, struct methodblock *);
JavaFrame *framePrev(JavaFrame *frame, JavaFrame *frame_buf);
static void compilerEnable();
static void compilerDisable();
static bool_t compileClass(ClassClass *);
static bool_t compileClasses(Hjava_lang_String *);
static HObject *compilerCommand(HObject *);

#define INITIALIZE_METHOD(MB) \
  {\
    (MB)->fb.access |= ACC_MACHINE_COMPILED;\
    (MB)->invoker = compileAndInvokeMethod;\
  }


/*
 * Initialization of the compiler.
 * This method is called by JVM.
 */
void java_lang_Compiler_start(void ** vector) {
  int version;
  int javaVersion, javaMinorVersion, compilerVersion;

#ifdef COMPILE_DEBUG
  printf("java_lang_Compiler_start() called.\n");
  fflush(stdout);
#endif

  /* get options */
  {
    char *opts = getenv("JAVA_COMPILER_OPT");
    char *opt;

    if (opts) {
#ifdef RUNTIME_DEBUG
      printf("JAVA_COMPILER_OPT: %s\n", opts);
      fflush(stdout);
#endif

      opt = strtok(opts, ", ");
      while (opt) {
	if (!strncmp(opt, "quiet", 6))
	  opt_quiet = 1;
	else if (!strncmp(opt, "outcode", 8))
	  opt_outcode = 1;
	else if (!strncmp(opt, "codesize", 9))
	  opt_codesize = 1;
	else if (!strncmp(opt, "dontcmplvmcls", 12))
	  opt_dontcmplvmcls = 1;
	opt = strtok(NULL, ", ");
      }
    }
  }


  /* show credit */
  if (!opt_quiet) {
    fprintf(stderr, CREDIT);
    fflush(stderr);
  }


  /* version check */

  version = *(int *)vector[0];
  compilerVersion = version >> 24;
  javaVersion = (version >> 16) & 0xff;
  javaMinorVersion = version & 0xffff;

  if (compilerVersion != COMPILER_VERSION) {
    printf("warning: version num. of compiler interface is not %d: %d\n",
		COMPILER_VERSION, compilerVersion);
  }
  if ( (javaVersion != JAVA_VERSION) ||
       (javaMinorVersion != JAVA_MINOR_VERSION) ) {
    printf("warning: version num. of class file format is not %d.%d: %d.%d\n",
	JAVA_VERSION, JAVA_MINOR_VERSION, javaVersion, javaMinorVersion);
  }


  /* initialization */

#ifdef __FreeBSD__
  /* to get addresses of functions in lib[cm] with sysDynamicLink() */
  /* libc for __divdi3, moddi3, printf, fflush */
  if (!sysAddDLSegment("/usr/lib/libc.so.3.1")) {
    if (!sysAddDLSegment("/usr/lib/libc.so.3.0")) {
      printf("fatal: JIT compiler can't find libc.so.3.1 or 3.0.\n"
		"  You must edit compiler.c and re-compile"
		"  if you use libc on your system.\n");
      sysExit(1);
    }
  }
  /* libm for fmod */
  if (!sysAddDLSegment("/usr/lib/libm.so.2.0")) {
    printf("fatal: JIT compiler can't find libm.so.2.0.\n"
		"  You must edit compiler.c and re-compile"
		"  if you use libc on your system.\n");
    sysExit(1);
  }
#endif

#ifdef RESOLVE_SYMBOL_ON_CODE
  initFunctionSymbols();
#endif

  /* compile the classes which is already loaded */
  if (!opt_dontcmplvmcls) {
    ClassClass **clazzptr;
    int i;

    BINCLASS_LOCK();
    for (i = nbinclasses, clazzptr = binclasses; --i >= 0; clazzptr++) {
#ifdef COMPILE_DEBUG
      printf("classes already in VM: %s\n", cbName(*clazzptr));
      fflush(stdout);
#endif
      initializeClass(*clazzptr);
    }
    BINCLASS_UNLOCK();
  }

  UseLosslessQuickOpcodes = FALSE;	/* prohibit generating lossy opcodes */

  /* set up link vector */

  *(char **)vector[1] = (char *)initializeClass;
  *(char **)vector[2] = (char *)invokeCompiledMethod;
  *(char **)vector[3] = (char *)signalHandler;
  *(char **)vector[4] = (char *)freeClass;
  *(char **)vector[5] = (char *)compileClass;
  *(char **)vector[6] = (char *)compileClasses;
  *(char **)vector[7] = (char *)compilerEnable;
  *(char **)vector[8] = (char *)compilerDisable;
  *(char **)vector[10] = (char *)pcInCompiledCode;
  *(char **)vector[11] = (char *)compiledCodePC;
  *(char **)vector[70] = (char *)framePrev;


#ifdef COMPILE_DEBUG
  printf("java_lang_Compiler_start() done.\n");
  fflush(stdout);
#endif
}


/*
 * Local Functions
 */

/*
 * Initialization the compiler at the time a class is loaded.
 */
void initializeClass(ClassClass *cb) {
  struct methodblock *mb, *mb_end;
#ifdef COMPILE_DEBUG
  printf("initializeClass(%s) called.\n", cbName(cb));
  fflush(stdout);
#endif

  if (!compiler_enabled) {
#ifdef COMPILE_DEBUG
    printf("  compiler has been disabled.\n  return.\n");
    fflush(stdout);
#endif
    return;
  }

  mb = cbMethods(cb);
  mb_end = mb + cbMethodsCount(cb);
  for (; mb < mb_end; mb++) {
    if (mb->fb.access & (ACC_ABSTRACT | ACC_NATIVE))
	/* methods not to be compiled */
	continue;
    INITIALIZE_METHOD(mb);
  }
}


static void signalHandler(int sig, void *info, void *uc) {
  printf("Signal %d is caught.\n", sig);

  showStackFrames(EE());
}


/*
 * Freeing class staffs related to the compiler.
 */
void freeClass(ClassClass *cb) {
  struct methodblock *mb = cbMethods(cb);
  struct methodblock *mb_end = mb + cbMethodsCount(cb);

#ifdef COMPILE_DEBUG
  printf("freeClass() called.\n");
  fflush(stdout);
#endif

  for (; mb < mb_end; mb++)
    freeMethod(mb);
}


unsigned char *compiledCodePC(JavaFrame *frame, struct methodblock *mb) {
  return (frame ? frame->lastpc : mb->code);
}


bool_t pcInCompiledCode(unsigned char *pc, struct methodblock *mb) {
  unsigned int off = pc - mb->code;
  return (off < mb->code_length);
}


JavaFrame *framePrev(JavaFrame *frame, JavaFrame *frame_buf) {
  return frame->prev;
}


/*
 * Enable the compiler.
 */
static void compilerEnable() {
  compiler_enabled = TRUE;
}


/*
 * Disable the compiler.
 */
static void compilerDisable() {
  compiler_enabled = FALSE;
}


/*
 * Compile the class.
 */
static bool_t compileClass(ClassClass *cb) {
  struct methodblock *mb, *mb_end;

#ifdef COMPILE_DEBUG
  printf("\n");
  printf("compileClass(%s) called.\n", cbName(cb));
  fflush(stdout);
#endif

  if (!compiler_enabled) {
#ifdef COMPILE_DEBUG
    printf("  compiler has been disabled.\n  return.\n");
    fflush(stdout);
#endif
    return TRUE;
  }

  mb = cbMethods(cb);
  mb_end = mb + cbMethodsCount(cb);
  for (; mb < mb_end; mb++) {
    if (!(mb->code)) {
#ifdef COMPILE_DEBUG
      printf("code is empty: %s %s\n", mb->fb.name, mb->fb.signature);
      fflush(stdout);
#endif
      INITIALIZE_METHOD(mb);
      continue;
    }
    if (compileMethod(mb))  return FALSE;
  }

#ifdef COMPILE_DEBUG
  printf("compileClass(%s) done.\n", cbName(cb));
  fflush(stdout);
#endif

  return TRUE;
}


/*
 * Compile the classes specified by comma-delimited names.
 */
static bool_t compileClasses(Hjava_lang_String *nameStr) {
  char *names = allocCString(nameStr);
  char *name = names;
  char *p = names;
  bool_t done = FALSE;

#ifdef COMPILE_DEBUG
  printf("compileClasses() called.\n");
  fflush(stdout);
#endif

  if (names) {
    ClassClass *cb;

    do {
      while ((*p != ',') && (*p != '\0'))  p++;
      if (*p == '\0')  done = TRUE;
      else  *p = '\0';

#ifdef COMPILE_DEBUG
      printf("compileClasses(): classname is `%s'\n", name);
      fflush(stdout);
#endif
      cb = FindClass(NULL, name, TRUE);
      if (cb)
	if (!(compileClass(cb)))  goto compileDone;

      p++;
      name = p;
    } while (!done);

  compileDone:
    sysFree(names);

    return TRUE;
  }

  return FALSE;
}


/*
 * Process the command from Java code.
 */
static HObject *compilerCommand(HObject *obj) {
  HObject *ret = NULL;
  
#ifdef COMPILE_DEBUG
  printf("compilerCommand() called.\n");
  fflush(stdout);
#endif

  /* do nothing. */

  return ret;
}
